/**************************************************************
ForwardAlgorithm.C
bmajoros@duke.edu

This is OPEN SOURCE SOFTWARE governed by the ARTISTIC LICENSE.
***************************************************************/
#include "BOOM/Constants.H"
#include <math.h>
#include "BOOM/String.H"
#include "ForwardAlgorithm.H"
#include <iostream>
using namespace std;

class NegativeInfinityException;

ForwardAlgorithm::ForwardAlgorithm(HiddenMarkovModel &hmm,
				   Sequence &sequence)
  : scalingFactors(sequence.getLength()+1),sequence(sequence),
    dpMatrix(hmm.countStates(),sequence.getLength()+1),
    numStates(hmm.countStates()),hmm(hmm),seqLen(sequence.getLength())
{
  computeDPMatrix();
}



double ForwardAlgorithm::getLogP()
{
  double sum=log(P);
  for(int i=1 ; i<=seqLen ; ++i) sum+=log(scalingFactors[i]);
  return sum;
}



double ForwardAlgorithm::getScaledP()
{
  return P;
}



DblArray1D &ForwardAlgorithm::getScalingFactors()
{
  return scalingFactors;
}



double ForwardAlgorithm::operator()(int state,int pos)
{
  return dpMatrix[state][pos];
}



double ForwardAlgorithm::computeScalingFactor(int i)
{
  double scalingFactor=0, sum;
  for(int l=1 ; l<numStates ; ++l)
    {
      sum=0;
      for(int k=0 ; k<numStates ; ++k)
	sum+=hmm.getTransitionProb(k,l)*dpMatrix[k][i-1];
      scalingFactor+=sum*hmm.getEmissionProb(l,sequence[i-1]);
    }
  if(isNaN(scalingFactor) || isInfinity(scalingFactor))
    throw String("scaling value in forward algorithm = ")+scalingFactor;
  /*if(scalingFactor<0 || scalingFactor>1)
    cerr << "scaling value in forward algorithm = "<<scalingFactor<<endl;*/
  return scalingFactor;
}



bool ForwardAlgorithm::PValueExceeds(ForwardAlgorithm &f)
{
  double difference=log(P)-log(f.P);
  for(int i=1 ; i<=seqLen ; ++i)
    {
      double d=log(scalingFactors[i])-log(f.scalingFactors[i]);
      difference+=d;
    }
  if(isInfinity(difference)||isNaN(difference)) 
    cerr << "ForwardAlgorithm: difference=" << difference << endl;
  bool exceeds=(difference>0);
  return exceeds;
}



void ForwardAlgorithm::computeDPMatrix()
{
  P=0;
  double s_sub_i, sum;
  dpMatrix[0][0]=1;
  for(int i=1 ; i<=seqLen ; ++i) dpMatrix[0][i]=0;
  for(int k=1 ; k<numStates ; ++k) dpMatrix[k][0]=0;
  for(int i=1 ; i<=seqLen ; ++i)
    {
      scalingFactors[i]=s_sub_i=computeScalingFactor(i);
      if(s_sub_i<=0) throw NegativeInfinityException();
      for(int l=1 ; l<numStates ; ++l)
	{
	  sum=0;	
	  for(int k=0 ; k<numStates ; ++k) 
	    sum+=hmm.getTransitionProb(k,l)*dpMatrix[k][i-1];
	  double v=dpMatrix[l][i]=sum*(hmm.getEmissionProb(l,sequence[i-1])
				       /s_sub_i);
          //cout<<v<<endl;
          if(isInfinity(v)) throw NegativeInfinityException();
	  if(v<0 || v>1.000001 || isNaN(v) || isInfinity(v))
          {
              cout<<sum<<" "<<hmm.getEmissionProb(l,sequence[i-1])<<" "<<s_sub_i<<endl;
              if(v<0) cout<<v<<"<0"<<endl;
              if(v>1) cout<<v<<">1"<<endl;
              if(isNaN(v)) cout<<v<<" isNaN"<<endl;
              if(isInfinity(v)) cout<<v<<" isInf"<<endl;
	    throw String("+cell (")+l+","+i+")="+v;
          }
	}
    }
  for(int k=1 ; k<numStates ; ++k) 
    P+=hmm.getTransitionProb(k,0)*dpMatrix[k][seqLen];
  if(P==0)
    for(int k=1 ; k<numStates ; ++k)
      {
	cout<<"dpMatrix[k][seqLen]="<<dpMatrix[k][seqLen]<<endl;
	cout<<"hmm.getTransitionProb(k,0)="
	    <<hmm.getTransitionProb(k,0)<<endl;
      }
}



